\ ARM Forth system
\ Reuben Thomas   11/3-27/10/99

\ Note: we assume that 1 CHARS = 1; to make this explicit, no conversion is
\ used. If chars were a different width, the dictionary encoding wouldn't
\ work anyway. 1 CELLS = 4 is also assumed, but the conversion words are
\ still used for convenience.

\ The current implementation of (CREATE) that doesn't alter T is often
\ assumed when calling CREATEd words


ONLY DEFINITIONS

INCLUDE ^.file/fs
INCLUDE ^.armasm/fs

VOCABULARY MFORTH
ONLY DEFINITIONS  ALSO ASSEMBLER
HEX



\ Registers

0F CONSTANT P  0E CONSTANT L  0D CONSTANT R  0C CONSTANT S  0B CONSTANT T
0A CONSTANT A  09 CONSTANT O



\ Inline words for ARM assembler

ALSO MFORTH DEFINITIONS

: Oplus   O O 800000 # ORR, ;
: minusO   O O 1000000 # ORR, ;
: plusO   O O 1800000 # ORR, ;
: minusOupdate   O O 1200000 # ORR, ;
: plusOupdate   O O 1A00000 # ORR, ;

: cr   SWI," OS_NewLine" ;
: space   100 BL + SWI, ;

: >name   T T 4 # ADD, ;
: >code   T T 0C # ADD, ;

: reset   O E0000000 # MOV, ;
: reverse   O O 10000000 # EOR, ;
: imm   O O 2000000 # EOR, ;
: set   O O 100000 # ORR, ;

PREVIOUS DEFINITIONS



\ Metacompiler for Machine Forth image

VOCABULARY META  ALSO META DEFINITIONS  ALSO MFORTH
HEX

0 VALUE IMAGE         \ address of image
0 VALUE LENGTH        \ length of image
0 VALUE THREADS       \ address of threads table
0 VALUE '(CREATE)     \ address of CREATE's runtime
0 VALUE '(CONSTANT)   \ address of CONSTANT's runtime
0 VALUE '(OP)         \ address of OP's runtime
0 VALUE '(COND)       \ address of COND's runtime
0 VALUE 'COMPILE,     \ address of COMPILE,
0 VALUE 'INLINE       \ address of INLINE

8000 CONSTANT ORIGIN
100 CONSTANT TIB-SIZE

: <ORG   IMAGE - ORIGIN + ;

VARIABLE CODE-CELLS  0 CODE-CELLS !
R: CODE,   ALIGN  ,  HERE CELL- DUP SYNCHRONIZE  1 CODE-CELLS +! ;
1 REDEFINER >SWITCH<


\ Navigate definitions; handle is address of thread field

: >NAME   ( thread-field -- name-field )   CELL+ ;
: >CODE   ( thread-field -- code-field )   3 CELLS + ;
: <COUNT   ( code-field -- count-field )   CELL- ;


\ Hash table and threads

10 CONSTANT #THREADS
: THREAD   ( name -- 'thread )   DUP 3 AND 2 LSHIFT  SWAP 4 RSHIFT 30 AND
   OR  THREADS + ;
: FIND   ( name -- xt TRUE | name FALSE )
   DUP THREAD                        \ get start of thread
   BEGIN @ DUP WHILE                 \ chain down the thread
      2DUP >NAME @ = IF              \ if name is the one we want
         NIP >CODE TRUE EXIT         \ leave xt and TRUE
      THEN
   REPEAT ;                          \ end-of-thread zero acts as FALSE
: RELOCATE
   THREADS  #THREADS CELLS  OVER + SWAP DO
      I
      BEGIN DUP @ ?DUP WHILE
         TUCK
         <M0 SWAP !
      REPEAT
      DROP
   CELL +LOOP ;


\ Definitions

: NAME   ( name )   ( -- name )
   BL WORD  DUP C@ 3 MIN             \ parse name
   1+ 3 LSHIFT  -1 SWAP LSHIFT  INVERT   \ make mask
   SWAP @ AND ;                      \ get name word and zero unused bytes
: '   NAME FIND 0= ABORT" not found in MF dictionary" ;
: $   ' BL, ;

: REG-SET   ( u -- mask )   1 SWAP LSHIFT 1- ;
: SAVE   ( mask -- )   R @! SWAP DB STM, ;
: RESTORE   ( mask -- )   R @! SWAP IA LDM, ;
: SAVE,   REG-SET SAVE ;
: RESTORE,   REG-SET RESTORE ;
: RET,   ?DUP IF  RESTORE,  THEN  RET, ;
: LINK,   REG-SET  [ 1 L LSHIFT ] LITERAL OR SAVE ;
: UNLINK,   REG-SET  [ 1 P LSHIFT ] LITERAL OR ^ RESTORE ;
: TAIL,   REG-SET  [ 1 L LSHIFT ] LITERAL OR RESTORE  ' B, ;

VARIABLE LAST
VARIABLE WORDS  0 WORDS !
: M:   ( -- count-field )   ( name )
   ALIGN  NAME                       \ parse name
   HERE  OVER THREAD  DUP @ ,  !     \ update thread
   ,                                 \ write name
   HERE  0 ,                         \ write count field
   HERE LAST !
   1 WORDS +! ;
: M;   DROP ;
: CREATE   ( name )   M: DROP  0 LINK,  '(CREATE) BL, ;
: CONSTANT   ( name )   M: DROP  0 LINK,  '(CONSTANT) BL,  , ;
: INLINE;   HERE OVER -  CELL-  SWAP C!  0 RET, ;
: IMMEDIATE   1  LAST @ <COUNT 3 + C! ;

: OP   ( name )   M: DROP  0 LINK,  '(OP) BL,  15 LSHIFT , ;
: COND   ( name )   M: DROP  0 LINK,  '(COND) BL,  1C LSHIFT , ;

: POSTPONE   ( name )   '  T S PUSH,  DUP <COUNT @ IF  <ORG T SWAP #MOV,
   'INLINE  ELSE <ORG T SWAP #MOV,  'COMPILE,  THEN  BL, ;



\ Build Machine Forth image

HERE TO IMAGE   \ note start address


\ Initialisation code

RESET   \ reset the assembler
>SWITCH<   \ switch in redefinitions
SWI," OS_GetEnv"   \ get end of memory in R1
R 1 TIB-SIZE # SUB,   \ set return stack to bottom of TIB
R HERE TUCK 8 + PCR STR,   \ store start of TIB (address resolved later)
S R 400 CELLS # SUB,   \ allow 1024 cells for return stack
AHEAD   \ branch to interpret


\ Allocate and zero the threads table

HERE DUP TO THREADS  #THREADS CELLS ALLOT  HERE OVER - ERASE


\ Runtimes

M: (CREATE)   HERE TO '(CREATE)
  A L FC000000 # BIC,
  0 UNLINK, M;
M: (CONSTANT)   HERE TO '(CONSTANT)
  A L FC000000 # BIC,
  T S PUSH,
  T A 0@ LDR,
  0 UNLINK, M;
M: (COND)   HERE TO '(COND)
  A L FC000000 # BIC,
  A A 0@ LDR,
  O O F0000000 # BIC,
  O O A ORR,
  0 UNLINK, M;
M: (LITERAL)
  T S PUSH,
  T L FC000000 # BIC,
  T T 0@ LDR,
  P L 4 # ADD, M;


\ Division

M: U/MOD
3 SAVE,
0 S 0@ LDR,                      \ get dividend
1 T MOV,                         \ copy divisor
BEGIN,                           \ shift divisor left as far as possible
   1 0 1 #LSR CMP,               \ without exceeding dividend
   1 1 1 #LSL LS MOV,
HI UNTIL,
2 0 # MOV,                       \ quotient=0
BEGIN,
   0 1 CMP,                      \ if dividend>divisor
   0 0 1 HS SUB,                 \ then dividend-=divisor
   2 2 2 ADC,                    \ shift quotient and add carry
   1 1 1 #LSR MOV,               \ shift divisor
   1 T CMP,                      \ continue until
LO UNTIL,                        \ divisor<original
0 S 0@ STR,                      \ store remainder
T 2 MOV,                         \ get quotient
3 RET,
M;

M: /MOD
5 SAVE,
4 L MOV,                         \ save link
0 S 0@ LDR,                      \ get dividend
3 0 T EOR,                       \ get signs of quotient
3 3 80000000 # AND,
3 3 0 1 #LSR ADD,                \ and remainder (=dividend's)
T 0 # CMP,                       \ ensure divisor is positive
T T 0 # MI RSB,
0 0 # CMP,                       \ ensure dividend is positive
0 0 0 # MI RSB,
0 S 0@ MI STR,                   \ store dividend if sign changed
$ U/MOD                          \ skip first instruction
3 80000000 # TST,                \ replace sign on quotient
T T 0 # NE RSB,
3 40000000 # TST,                \ replace sign on remainder
0 0 0 # NE RSB,
0 S 0@ NE STR,                   \ push remainder if necessary
L 4 MOV,
5 RESTORE,
0 RET,
M;


\ Inline ops

M: DUP   T S PUSH, INLINE;
M: OVER   T S PUSH,  T S 4 #+@ LDR, INLINE;
M: DROP   T S POP, INLINE;

M: INVERT   T T MVN, INLINE;
M: AND   0 S POP,  T 0 T AND, INLINE;
M: XOR   0 S POP,  T 0 T EOR, INLINE;
M: +   0 S POP,  T 0 T ADD, INLINE;
M: -   0 S POP,  T 0 T SUB, INLINE;
M: *   0 S POP,  T 0 T MUL, INLINE;


\ Utility words

M: EXECUTE   0 T MOV,  T S POP,  P 0 MOV, M;
M: SWAP   0 S POP,  T S PUSH,  T 0 MOV,  0 RET, M;

M: CR   SWI," OS_NewLine" INLINE;
M: SPACE   100 BL + SWI, INLINE;
M: EMIT   0 T MOV,  SWI," OS_WriteC"  T S POP,  0 RET, M;
M: BYE   cr  0 0 # MOV,  SWI," OS_Exit" M;


\ Number I/O

CREATE BASE  0A ,
M: >DIGIT   ( n -- char )
  T A # CMP,
  T T CHAR 0 LO # ADD,
  T T CHAR A 0A - GT HS # ADD,
  0 RET, M;
M: <DIGIT   ( char -- n )
  T CHAR A # CMP,
  T T CHAR 0 LO # SUB,
  T T CHAR A 0A - HS # SUB,
  0 RET, M;

M: U.
  1 LINK,
  0 0 # MVN,
  0 S PUSH,
  $ BASE
  0 A 0@ LDR,
  BEGIN,
    T S PUSH,
    T 0 MOV,
    $ U/MOD
    T 0 # CMP,
  EQ UNTIL,
  T S POP,
  BEGIN,
    $ >DIGIT  $ EMIT
    T 0 # CMP,
  LT UNTIL,
  T S POP,
  space
  1 UNLINK, M;
M: .   ( n -- )
  T 0 # CMP,
  100 CHAR - + MI SWI,
  T T 0 MI # RSB,
  ' U. B, M;

M: SIGN   ( start -- sign start' )
  1 SAVE,
  A T 0@ BYTE LDR,
  A CHAR - # CMP,
  T T 1 EQ # ADD,
  0 1 NE # MOV,
  0 0 EQ # MVN,
  0 S PUSH,
  1 RET, M;
M: DIGIT?   ( n -- f )
  0 LINK,
  $ BASE
  A A 0@ LDR,
  T A CMP,
  T 0 LO # MVN,
  T 0 HS # MOV,
  0 UNLINK, M;
M: NUMBER   ( start length -- n f )
  5 LINK,
  0 T MOV,
  T S POP,
  0 0 T ADD,
  $ SIGN
  1 0 # MOV,
  2 T MOV,
  $ BASE
  3 A 0@ LDR,
  BEGIN,
    1 3 1 MUL,
    T 2 1 #@+ BYTE LDR,
    $ <DIGIT
    4 T MOV,
    $ DIGIT?
    T 0 # CMP,
    NE IF,
      1 1 4 ADD,
      2 0 CMP,
  SWAP EQ UNTIL,
      0 S POP,
      1 0 1 MUL,
      1 S PUSH,
      T 0 # MVN,
      5 UNLINK,
    THEN,
    T 0 # MOV,
    5 UNLINK, M;


\ Terminal I/O

M: TYPE
  1 SAVE,
  A S POP,
  BEGIN,
    0 A 1 #@+ BYTE LDR,
    SWI," OS_WriteC"
    T T 1 SET # SUB,
  EQ UNTIL,
  T S POP,
  1 RET, M;

CREATE TIB  SWAP DUP HERE >-< 8 -  OVER @ OR  SWAP !  0 ,
CREATE #TIB  0 ,
CREATE >IN  0 ,
M: REFILL
  4 LINK,
  $ TIB
  0 A 0@ LDR,
  $ >IN
  0 A 0@ STR,
  1 TIB-SIZE # MOV,
  2 BL # MOV,
  3 FF # MOV,
  SWI," OS_ReadLine"
  $ #TIB
  1 A 0@ STR,
  4 UNLINK, M;

M: SKIP   ( delim ptr -- ptr' )
  1 SAVE,
  A S POP,
  BEGIN,
    0 T 1 #@+ BYTE LDR,
    0 A CMP,
  NE UNTIL,
  T T 1 # SUB,
  1 RET, M;
M: PARSE   ( delim ptr -- ptr' )
  1 SAVE,
  A S POP,
  BEGIN,
    0 T 1 #@+ BYTE LDR,
    0 A CMP,
  EQ UNTIL,
  T T 1 # SUB,
  1 RET, M;
M: WORD   ( char -- start len )
  3 LINK,
  $ TIB
  0 A 0@ LDR,
  $ #TIB
  1 A 0@ LDR,
  0 0 1 ADD,
  1 T MOV,
  T S PUSH,
  T 0A # MOV,
  T 0 0@ BYTE STR,
  $ >IN
  2 A MOV,
  T A 0@ LDR,
  $ SKIP
  1 0 0@ BYTE STR,
  1 S PUSH,
  0 T MOV,
  $ PARSE
  T 2 0@ STR,
  0 S PUSH,
  T T 0 SUB,
  3 UNLINK, M;


\ Dictionary

M: >NAME   T T 4 # ADD, INLINE;
M: >CODE   T T 0C # ADD, INLINE;
CREATE THREADS  THREADS <ORG ,
M: THREAD   ( name -- 'thread )
  0 LINK,
  A T 3 # AND,
  T T 300 # AND,
  T T 4 #LSR MOV,
  T T A 2 #LSL ORR,
  $ THREADS
  A A 0@ LDR,
  T T A ADD,
  0 UNLINK, M;
M: NAME   ( start length -- name )
  2 SAVE,
  A T MOV,
  A 3 # CMP,
  A 3 # HI MOV,
  A A 3 #LSL MOV,
  1 S POP,
  1 1 1 # SUB,
  BEGIN,
    0 1 A 3 #LSR +@ BYTE LDR,
    T T 0 A LSL ORR,
    A A 8 SET # SUB,
  EQ UNTIL,
  2 RET, M;
M: FIND   ( name -- xt | FALSE )
  2 LINK,
  0 T MOV,
  $ THREAD
  BEGIN,
    T T 0@ LDR,
    T 0 # CMP,
    NE WHILE,
    1 T MOV,
    >name
    T T 0@ LDR,
    T 0 CMP,
    T 1 MOV,
    NE IF,
  -ROT REPEAT,
      2 UNLINK,
    THEN,
    >code
    2 UNLINK, M;
M: '
  0 LINK,
  T S PUSH,
  T BL # MOV,
  $ WORD
  $ NAME
  0 TAIL, FIND M;


\ Compiler #1

CREATE DP  HERE 0 ,
CREATE STATE  1 ,   \ N.B.  0 = compiling, 1 = interpreting

M: SYNCHRONIZE   ( a-addr -- )
   3 SAVE,
   0 1 # MOV,
   1 T MOV,
   2 T MOV,
   SWI," XOS_SynchroniseCodeAreas"
   3 RET, M;
M: CODE!   ( x -- )   ( address in A )
  3 SAVE,
  T A 0@ STR,
  T S POP,
  0 1 # MOV,
  1 A MOV,
  2 A MOV,
  SWI," XOS_SynchroniseCodeAreas"
  3 RET, M;
M: ,
  2 LINK,
  $ DP
  0 A 0@ LDR,
  1 0 4 # ADD,
  1 A 0@ STR,
  A 0 MOV,
  2 TAIL, CODE! M;
M: >BRANCH   ( from to -- offset )
  A S POP,
  T T A SUB,
  T T 2 #LSR MOV,
  T T 2 # SUB,
  T T FF000000 # BIC,
  0 RET, M;


\ ARM assembler

0F CONSTANT P  0E CONSTANT L  0D CONSTANT R  0C CONSTANT S  0B CONSTANT T
0A CONSTANT A  09 CONSTANT O

M: MINIMIZE   ( x1 -- x2 u )   \ minimize unsigned value of x1 by rotation
   2 SAVE,
   T 100 # CMP,                      \ stop if already a usable constant
   T S LO PUSH,                      \ pushing 0 as the rotation
   T 0 # LO MOV,
   2 LO RESTORE,                     \ can't use 2 LO RET, as condition code
   0 LO RET,                         \ is not used for return instruction
   0 T MOV,                          \ hold minimum value found so far
   A F # MOV,                        \ current rotation
   1 0 # MOV,                        \ rotation for minimum value
   BEGIN,                            \ try each rotation
     0 0 1E #ROR MOV,
     T 0 CMP,
     T 0 HI MOV,
     1 A HI MOV,
     A A 1 SET # SUB,
   EQ UNTIL,
   1 S PUSH,                         \ push new rotation
   2 RET, M;

00 COND EQ  01 COND NE  02 COND CS  03 COND CC  04 COND MI  05 COND PL
06 COND VS  07 COND VC  08 COND HI  09 COND LS  0A COND GE  0B COND LT
0C COND GT  0D COND LE  0E COND AL  03 COND LO  02 COND HS

M: REVERSE   O O 10000000 # EOR, INLINE;
M: RESET   O E0000000 # MOV, INLINE;

M: OP,
  1 LINK,
  S @! 401 IA LDM,   \ load R0 and A
  O O T ORR,
  O O 0 10 #LSL ORR,
  T O A 0C #LSL ORR,
  $ ,
  reset
  1 UNLINK, M;
M: (OP)   HERE TO '(OP)
  A L FC000000 # BIC,
  A A 0@ LDR,
  O O A ORR,
  0 TAIL, OP, M;
00 OP AND,  01 OP EOR,  02 OP SUB,  03 OP RSB,  04 OP ADD,  05 OP ADC,
06 OP SBC,  07 OP RSC,  08 OP xx1   09 OP xx2   0A OP xx3   0B OP xx4
0C OP ORR,  0D OP xx5   0E OP BIC,  0F OP xx6

M: SET   O O 100000 # ORR, INLINE;

M: TST,
  1 SAVE,
  0 S POP,
  A 0 # MOV,
  S @! 401 DB STM,   \ store R0 and A
  set
  1 RESTORE,
  ' xx1 B, M;
M: TEQ,
  1 SAVE,
  0 S POP,
  A 0 # MOV,
  S @! 401 DB STM,   \ store R0 and A
  set
  1 RESTORE,
  ' xx2 B, M;
M: CMP,
  1 SAVE,
  0 S POP,
  A 0 # MOV,
  S @! 401 DB STM,   \ store R0 and A
  set
  1 RESTORE,
  ' xx3 B, M;
M: CMN,
  1 SAVE,
  0 S POP,
  A 0 # MOV,
  S @! 401 DB STM,   \ store R0 and A
  set
  1 RESTORE,
  ' xx4 B, M;
M: MOV,
  A 0 # MOV,
  A S PUSH,
  ' xx5 B, M;
M: MVN,
  A 0 # MOV,
  A S PUSH,
  ' xx6 B, M;

M: MUL,
  3 LINK,
  S @! 403 IA LDM,   \ load R0-R1 and A
  T T 8 #LSL MOV,
  O O T ORR,
  O O 90 # ORR,
  O O 0 ORR,
  O O 1 ORR,
  O O A 10 #LSL ORR,
  $ ,
  reset
  3 UNLINK, M;
M: MLA,
  T T 0C #LSL MOV,
  T T 90 # ORR,
  T T 200000 # ORR,
  O O T ORR,
  ' MUL, B, M;

M: LSL
  T T 8 #LSL MOV,
  T T 10 # ORR,
  O O T ORR,
  T S POP,
  0 RET, M;
M: #LL
  T T 7 #LSL MOV,
  O O T ORR,
  T S POP,
  0 RET, M;
M: LSR
  T T 8 #LSL MOV,
  T T 30 # ORR,
  O O T ORR,
  T S POP,
  0 RET, M;
M: #LR
  T T 7 #LSL MOV,
  T T 20 # ORR,
  O O T ORR,
  T S POP,
  0 RET, M;
M: ASR
  T T 8 #LSL MOV,
  T T 50 # ORR,
  O O T ORR,
  T S POP,
  0 RET, M;
M: #AR
  T T 7 #LSL MOV,
  T T 40 # ORR,
  O O T ORR,
  T S POP,
  0 RET, M;
M: ROR
  T T 8 #LSL MOV,
  T T 70 # ORR,
  O O T ORR,
  T S POP,
  0 RET, M;
M: #RR
  T T 7 #LSL MOV,
  T T 60 # ORR,
  O O T ORR,
  T S POP,
  0 RET, M;
M: RRX
  T S PUSH,
  T 0 # MOV,
  ' #RR B, M;

M: #
  0 LINK,
  $ MINIMIZE
  T T 8 #LSL MOV,
  O O T ORR,
  O O 2000000 # ORR,
  T S POP,
  0 UNLINK, M;

M: (B)
  1 LINK,
  0 S POP,
  $ DP
  T S PUSH,
  A A 0@ LDR,
  A S PUSH,
  T 0 MOV,
  $ >BRANCH
  A S POP,
  O O A ORR,
  T O T ORR,
  $ ,
  reset
  1 UNLINK, M;
M: B,
  T S PUSH,
  T A000000 # MOV,
  ' (B) B, M;
M: BL,
  T S PUSH,
  T B000000 # MOV,
  ' (B) B, M;

M: @- M;
M: @+   O O 800000 # ORR, INLINE;
M: -@   O O 1000000 # ORR, INLINE;
M: +@   O O 1800000 # ORR, INLINE;
M: -@!   O O 1200000 # ORR, INLINE;
M: +@!   O O 1A00000 # ORR, INLINE;

M: OFFSET
  T T 2000000 # ORR,
  O O T ORR,
  T 0 # MOV,
  0 RET, M;
M: #@-
  ' OFFSET B, M;
M: #@+
  0 LINK,
  $ OFFSET
  Oplus
  0 UNLINK, M;
M: #-@
  0 LINK,
  $ OFFSET
  minusO
  0 UNLINK, M;
M: #+@
  0 LINK,
  $ OFFSET
  plusO
  0 UNLINK, M;
M: #-@!
  0 LINK,
  $ OFFSET
  minusOupdate
  0 UNLINK, M;
M: #+@!
  0 LINK,
  $ OFFSET
  plusOupdate
  0 UNLINK, M;
M: 0@
  T S PUSH,
  T 0 # MOV,
  ' #+@ B, M;

M: IMM   O O 2000000 # EOR, INLINE;
M: BYTE   O O 400000 # ORR, INLINE;
M: LDR,   imm  O O 4100000 # ORR,  ' OP, B, M;
M: STR,   imm  O O 4000000 # ORR,  ' OP, B, M;

M: PCR
  0 LINK,
  $ DP
  A A 0@ LDR,
  A A 8 # ADD,
  T T A SET SUB,
  T T 0 # MI RSB,
  MI $ #-@
  PL $ #+@
  A 0F # MOV,
  A S PUSH,
  0 UNLINK, M;

0           CONSTANT DA  1 17 LSHIFT CONSTANT IA
2 17 LSHIFT CONSTANT DB  3 17 LSHIFT CONSTANT IB

M: @!   O O 200000 # ORR, INLINE;
M: ^   O O 400000 # ORR, INLINE;
M: (LDSTM)
  2 LINK,
  T T O ORR,
  S @! 403 IA LDM,   \ load R0-R1 and A
  T T 0 ORR,
  T T 1 ORR,
  T T A 10 #LSL ORR,
  $ ,
  reset
  2 UNLINK, M;
M: LDM,
  T S PUSH,
  T 8100000 # MOV,
  ' (LDSTM) B, M;
M: STM,
  T S PUSH,
  T 8000000 # MOV,
  ' (LDSTM) B, M;

M: SWI,
  0 LINK,
  O O F000000 # ORR,
  T T O ORR,
  $ ,
  reset
  0 UNLINK, M;
M: X   O O 20000 # ORR, INLINE;

M: AHEAD,
  0 LINK,
  T S PUSH,
  $ DP
  T A 0@ LDR,
  T S PUSH,
  T T 8 # ADD,
  0 TAIL, B, M;
M: IF,
  reverse
  ' AHEAD, B, M; IMMEDIATE
M: THEN,
  1 LINK,
  $ DP
  A A 0@ LDR,
  T S PUSH,
  0 T 0@ LDR,
  0 S PUSH,
  T S PUSH,
  T A MOV,
  $ >BRANCH
  A S POP,
  T T A ORR,
  A S POP,
  1 TAIL, CODE! M; IMMEDIATE
M: ELSE,
  0 LINK,
  $ AHEAD,
  $ SWAP
  0 TAIL, THEN, M; IMMEDIATE

M: BEGIN,
  0 LINK,
  T S PUSH,
  $ DP
  T A 0@ LDR,
  0 UNLINK, M; IMMEDIATE
M: AGAIN,
  ' B, B, M; IMMEDIATE
M: UNTIL,
  reverse
  ' B, B, M; IMMEDIATE
M: WHILE,
  0 LINK,
  $ IF,
  $ SWAP
  0 UNLINK, M; IMMEDIATE
M: REPEAT,
  0 LINK,
  $ AGAIN,
  0 TAIL, THEN, M; IMMEDIATE

M: RET,   0 LINK,  $ (LITERAL)  P L SET MOV,  $ ,  0 UNLINK, M;
M: LINK,   0 LINK,  $ (LITERAL)  L R PUSH,  $ ,  0 UNLINK, M;
M: UNLINK,   0 LINK,  $ (LITERAL)  R @! 8000 IA ^ LDM,  $ ,  0 UNLINK, M;


\ Compiler #2

M: INLINE   ( xt -- )   HERE TO 'INLINE
  2 LINK,
  0 T 4 # SUB,
  0 0 0@ BYTE LDR,
  1 T MOV,
  BEGIN,
    T 1 4 #@+ LDR,
    $ ,
    0 0 4 # SET SUB,
    T S NE PUSH,
  EQ UNTIL,
  2 UNLINK, M;
M: [
  1 LINK,
  $ STATE
  0 1 # MOV,
  0 A 0@ STR,
  1 UNLINK, M; IMMEDIATE
M: ]
  1 LINK,
  $ STATE
  0 0 # MOV,
  0 A 0@ STR,
  1 UNLINK, M;

M: >BRANCH   ( from to -- offset )
  A S POP,
  T T A SUB,
  T T 2 #LSR MOV,
  T T 2 # SUB,
  T T FF000000 # BIC,
  0 RET, M;
M: BRANCH   ( to opcode -- )
  1 LINK,
  0 T MOV,
  T S POP,
  $ DP
  A A 0@ LDR,
  A S PUSH,
  $ >BRANCH
  T T 0 ORR,
  1 TAIL, , M;
M: BRANCH,
  T S PUSH,
  T EA000000 # MOV,
  ' BRANCH B, M;
M: COMPILE,    HERE TO 'COMPILE,
  T S PUSH,
  T EB000000 # MOV,
  ' BRANCH B, M;

M: LITERAL
  0 LINK,
  POSTPONE (LITERAL)
  0 TAIL, , M;

M: CODE   ( name )
  2 LINK,
  T S PUSH,
  T BL # MOV,
  $ WORD
  $ NAME
  $ DP
  0 A 0@ LDR,
  T S PUSH,
  $ THREAD
  1 T MOV,
  T T 0@ LDR,
  $ ,
  0 1 0@ STR,
  $ ,
  T S PUSH,
  T 0 8 # ADD,
  T S PUSH,
  T 0 # MOV,
  $ ,
  reset
  2 UNLINK, M;
M: END-CODE
  T S POP,
  0 RET, M; IMMEDIATE
M: :   ( name )
  0 LINK,
  $ CODE
  0 TAIL, ] M;
M: ;
  0 LINK,
  $ UNLINK,
  $ END-CODE
  0 TAIL, [ M; IMMEDIATE
M: CREATE   ( name )
  0 LINK,
  $ CODE
  $ LINK,
  POSTPONE (CREATE)
  0 TAIL, END-CODE M;


\ Interpreter

M: COMPILE   ( xt -- )
  A T 4 # SUB,
  A A 0@ BYTE LDR,
  A 0 # CMP,
  ' COMPILE, EQ B,
  ' INLINE B, M;
M: INTERPRET
  2 LINK,
  BEGIN,
    T S PUSH,
    T BL # MOV,
    $ WORD
    T 0 # CMP,
    NE WHILE,
    0 T MOV,
    1 S 0@ LDR,
    $ NAME
    $ FIND
    T 0 # CMP,
    NE IF,
      0 T 1 #-@ BYTE LDR,
      $ STATE
      1 A 0@ LDR,
      0 0 1 SET ORR,
      NE $ EXECUTE
      EQ $ COMPILE
    ELSE,
      1 S PUSH,
      T 0 MOV,
      $ NUMBER
      T 0 # CMP,
      EQ IF,
        100 CHAR ? + SWI,
        cr
        IMAGE B,
      THEN,
      T S POP,
      $ STATE
      0 A 0@ LDR,
      0 0 # CMP,
      EQ $ LITERAL
    THEN,
  REPEAT,
  S S 4 # ADD,
  T S POP,
  2 UNLINK, M;

M: QUIT
  IMAGE B,    \ branch to start of image to reset stacks
  ROT THEN,   \ resolve initial branch
  T 1 # MOV,
  $ STATE
  T A 0@ STR, \ reset STATE
  BEGIN,
    cr
    $ REFILL
    100 CHAR > + SWI,
    space
    $ INTERPRET
  AGAIN, M;

HERE <ORG  SWAP !



\ Relocate and save image

HERE IMAGE - TO LENGTH

M0  IMAGE TO M0  RELOCATE  TO M0   \ relocate image
S" armf"  IMAGE LENGTH 2OVER SAVE-FILE   \ write object module
FF8 -ROT TYPE-FILE   \ set filetype to Absolute



\ Print out statistics

DECIMAL
CR LENGTH CELL/ .  .( cells total )
CR CODE-CELLS ?  .( code cells )
CR WORDS ?  .( definitions )



>SWITCH<   \ switch back to original definitions
KERNEL PREVIOUS DEFINITIONS   \ restore original order
