\ Machine Forth assembler for ARM
\ Reuben Thomas
\ 11/3-5/4/99

CR .( Machine Forth assembler )


VOCABULARY MFORTH  ALSO MFORTH DEFINITIONS  ALSO ASSEMBLER
DECIMAL


\ Registers

10 CONSTANT T  11 CONSTANT A  12 CONSTANT S
13 CONSTANT R  14 CONSTANT L  15 CONSTANT P


\ Utility words for peepholing stack pops and pushes

VARIABLE DELTA  0 DELTA !   \ actual - logical stack depth

: LEVEL   ( delta -- )   ?DUP IF  S S  ROT CELLS #ADD,  THEN  0 DELTA ! ;
: POP-T   1 DELTA +! ;
: PUSH-T   -1 DELTA +! ;
: READ-T   DELTA @  ?DUP IF  1- LEVEL  T S POP,  THEN ;
: WRITE-T   PUSH-T  DELTA @  ?DUP IF  DUP 0> IF  LEVEL  ELSE DROP  T S PUSH,
   0 DELTA !  THEN  THEN ;


\ Counting the number of Machine Forth ops compiled

VARIABLE MF-OPS  0 MF-OPS !

: INC   1 MF-OPS +! ;
: I:   :  POSTPONE INC ;


\ Zero-parameter ops

I: ret   READ-T  P L SET MOV, ;
I: col   L R PUSH, ;
I: scol   READ-T  P R POP, ;
I: A@   WRITE-T  T A MOV, ;
I: A!   READ-T  A T MOV,  POP-T ;
I: @A   WRITE-T  T A 0@ LDR, ;
I: !A   READ-T  T A 0@ STR,  POP-T ;
I: B@A   WRITE-T  T A 0@ BYTE LDR, ;
I: B!A   READ-T  T A 0@ BYTE STR,  POP-T ;
I: @A+   WRITE-T  T A 4 #@+ LDR, ;
I: !A+   READ-T  T A 4 #@+ STR,  POP-T ;
I: B@A+   WRITE-T  T A 1 #@+ BYTE LDR, ;
I: B!A+   READ-T  T A 1 #@+ BYTE STR,  POP-T ;
I: pop   WRITE-T  T L MOV,  L R POP, ;
I: push   READ-T  L R PUSH,  L T MOV,  POP-T ;
I: @R+   WRITE-T  T L 4 #@+ LDR, ;
I: !R+   READ-T  T L 4 #@+ STR,  POP-T ;
I: B@R+   WRITE-T  T L 1 #@+ BYTE LDR, ;
I: B!R+   READ-T  T L 1 #@+ BYTE STR,  POP-T ;
I: com   READ-T  T T MVN, ;
I: 2mul   READ-T  T T 1 #ASL MOV, ;
I: 2div   READ-T  T T 1 #ASR MOV, ;
I: mul   READ-T  R0 S POP,  T R0 T MUL, ;
I: -or   READ-T  R0 S POP,  T T R0 EOR, ;
I: and   READ-T  R0 S POP,  T T R0 AND, ;
I: plus   READ-T  R0 S POP,  T T R0 ADD, ;
I: dup   READ-T  T S PUSH, ;
I: over   WRITE-T  T S 4 #+@ LDR, ;
I: drop   POP-T ;


\ Ops that take parameters

: ENUMERATE   0 ?DO  I  LOOP ;

I: lit   WRITE-T  T SWAP #MOV, ;
I: else   READ-T  B, ;
I: if   READ-T  T 0 # CMP,  EQ B, ;
I: -if   READ-T  T 1 31 LSHIFT # TST,  EQ B, ;
I: call   READ-T  BL, ;
I: swi   ( ins outs swi -- )
   READ-T
   T S PUSH,
   ROT ?DUP IF
      S @! { ROT ENUMERATE } FD LDM,
   THEN
   SWI,
   ?DUP IF
      S @! { ROT ENUMERATE } FD STM,
   THEN
   POP-T ;


\ Pseudo-ops

: ahead,   READ-T  HERE  DUP 8 + else ;
: if,   READ-T  HERE 4 +  DUP 8 + if ;
: -if,   READ-T  HERE 4 +  DUP 8 + -if ;
: then,   READ-T  DUP @  OVER HERE >BRANCH  OR  SWAP CODE! ;
: else,   ahead,  SWAP then, ;

: begin,   READ-T  HERE ;
: again,   else ;
: while,   if, SWAP ;
: -while,   -if, SWAP ;
: repeat,   again, then, ;
: until,   if ;
: -until,   -if ;
: 0until,   while, repeat, ;
: +until,   -while, repeat, ;

: swi"   ( name )   ( ins outs -- )   SWI" swi ;


PREVIOUS PREVIOUS