\ ARM Machine Forth system
\ Reuben Thomas   11/3-27/10/99

\ Note: we assume that 1 CHARS = 1; to make this explicit, no conversion is
\ used. If chars were a different width, the dictionary encoding wouldn't
\ work anyway. 1 CELLS = 4 is also assumed, but the conversion words are
\ still used for convenience.

ONLY DEFINITIONS

INCLUDE ^.file/fs
INCLUDE ^.armasm/fs
INCLUDE asm/fs

ONLY DEFINITIONS  ALSO ASSEMBLER



\ Extra ops for Machine Forth

ALSO MFORTH DEFINITIONS

I: negate   READ-T  T T 0 # RSB, ;
I: minus   READ-T  R0 S POP,  T R0 T SUB, ;
I: or   READ-T  R0 S POP,  T R0 T ORR, ;
I: lshift   READ-T  R0 S POP,  T R0 T LSL MOV, ;
I: rshift   READ-T  R0 S POP,  T R0 T LSR MOV, ;

PREVIOUS DEFINITIONS



\ Metacompiler for Machine Forth image

VOCABULARY META  ALSO META DEFINITIONS  ALSO MFORTH
DECIMAL

0 VALUE IMAGE       \ address of image
0 VALUE LENGTH      \ length of image
0 VALUE THREADS     \ address of threads table
0 VALUE '(CREATE)   \ address of CREATE's runtime
0 VALUE 'COMPILE,   \ address of COMPILE,
0 VALUE 'INLINE     \ address of INLINE

HEX 8000 DECIMAL CONSTANT ORIGIN
256 CONSTANT TIB-SIZE

: <ORG   IMAGE - ORIGIN + ;

VARIABLE CODE-CELLS  0 CODE-CELLS !
R: CODE,   ALIGN  ,  HERE CELL- DUP SYNCHRONIZE  1 CODE-CELLS +! ;
1 REDEFINER >SWITCH<


\ Navigate definitions; handle is address of thread field

: >NAME   ( thread-field -- name-field )   CELL+ ;
: >CODE   ( thread-field -- code-field )   3 CELLS + ;
: <COUNT   ( code-field -- count-field )   CELL- ;


\ Hash table and threads

16 CONSTANT #THREADS
: THREAD   ( name -- 'thread )   DUP 3 AND 2 LSHIFT  SWAP 4 RSHIFT 48 AND
   OR  THREADS + ;
: FIND   ( name -- xt TRUE | name FALSE )
   DUP THREAD                        \ get start of thread
   BEGIN @ DUP WHILE                 \ chain down the thread
      2DUP >NAME @ = IF              \ if name is the one we want
         NIP >CODE TRUE EXIT         \ leave xt and TRUE
      THEN
   REPEAT ;                          \ end-of-thread zero acts as FALSE
: RELOCATE
   THREADS  #THREADS CELLS  OVER + SWAP DO
      I
      BEGIN DUP @ ?DUP WHILE
         TUCK
         <M0 SWAP !
      REPEAT
      DROP
   CELL +LOOP ;


\ Definitions

: NAME   ( name )   ( -- name )
   BL WORD  DUP C@ 3 MIN             \ parse name
   1+ 3 LSHIFT  -1 SWAP LSHIFT  INVERT   \ make mask
   SWAP @ AND ;                      \ get name word and zero unused bytes
: '   NAME FIND 0= ABORT" not found in MF dictionary" ;
: $   ' call ;
: r$   col $ pop drop ;

VARIABLE LAST
VARIABLE WORDS  0 WORDS !
: M:   ( -- count-field )   ( name )
   0 DELTA !
   ALIGN  NAME                       \ parse name
   HERE  OVER THREAD  DUP @ ,  !     \ update thread
   ,                                 \ write name
   HERE  0 ,                         \ write count field
   HERE LAST !
   1 WORDS +! ;
: M;   DROP  0 DELTA ! ;
: CREATE   ( name )   M: DROP  col '(CREATE) call ;
: INLINE;   READ-T  HERE OVER -  CELL-  SWAP C!  ret ;
: IMMEDIATE   1  LAST @ <COUNT 3 + C! ;

: POSTPONE   ( name )   '  DUP <COUNT @ IF  <ORG lit  'INLINE call
   ELSE <ORG lit  'COMPILE, call  THEN ;



\ Build Machine Forth image

HERE TO IMAGE   \ note start address


\ Initialisation code

RESET   \ reset the assembler
>SWITCH<   \ switch in redefinitions
SWI," OS_GetEnv"   \ get end of memory in R1
R R1 TIB-SIZE # SUB,   \ set return stack to bottom of TIB
R HERE TUCK 8 + PCR STR,   \ store start of TIB (address resolved later)
S R 1024 CELLS # SUB,   \ allow 1024 cells for return stack
ahead,   \ branch to interpret


\ Allocate and zero the threads table

HERE DUP TO THREADS  #THREADS CELLS ALLOT  HERE OVER - ERASE


\ Runtimes

M: (CREATE)   HERE TO '(CREATE)  pop HEX 03FFFFFF DECIMAL lit and A!  ret M;


\ Division

HEX

M: U/MOD
R0 S 0@ LDR,                     \ get dividend
R1 T MOV,                        \ copy divisor
BEGIN,                           \ shift divisor left as far as possible
   R1 R0 1 #LSR CMP,             \ without exceeding dividend
   R1 R1 1 #LSL LS MOV,
HI UNTIL,
R2 0 # MOV,                      \ quotient=0
BEGIN,
   R0 R1 CMP,                    \ if dividend>divisor
   R0 R0 R1 HS SUB,              \ then dividend-=divisor
   R2 R2 R2 ADC,                 \ shift quotient and add carry
   R1 R1 1 #LSR MOV,             \ shift divisor
   R1 T CMP,                     \ continue until
LO UNTIL,                        \ divisor<original
R0 S 0@ STR,                     \ store remainder
T R2 MOV,                        \ get quotient
ret
M;

M: /MOD
R4 L MOV,                        \ save link
R0 S 0@ LDR,                     \ get dividend
R3 R0 T EOR,                     \ get signs of quotient
R3 R3 80000000 # AND,
R3 R3 R0 1 #LSR ADD,             \ and remainder (=dividend's)
T 0 # CMP,                       \ ensure divisor is positive
T T 0 # MI RSB,
R0 0 # CMP,                      \ ensure dividend is positive
R0 R0 0 # MI RSB,
' U/MOD CELL+ call               \ skip first instruction
R3 80000000 # TST,               \ replace sign on quotient
T T 0 # NE RSB,
R3 40000000 # TST,               \ replace sign on remainder
R0 R0 0 # NE RSB,
R0 S 0@ NE STR,                  \ push remainder if necessary
P R4 MOV,
M;

DECIMAL


\ MF inline ops

M: ret   ret INLINE;     M: !A+   !A+ INLINE;      M: 2*   2mul INLINE;
M: col   col INLINE;     M: B@A+   B@A+ INLINE;    M: 2/   2div INLINE;
M: scol   scol INLINE;   M: B!A+   B!A+ INLINE;    M: *   mul INLINE;
M: A@   A@ INLINE;       M: R>   pop INLINE;       M: XOR   -or INLINE;
M: A!   A! INLINE;       M: >R   push INLINE;      M: AND   and INLINE;
M: @A   @A INLINE;       M: @R+   @R+ INLINE;      M: +   plus INLINE;
M: !A   !A INLINE;       M: !R+   !R+ INLINE;      M: DUP   dup INLINE;
M: B@A   B@A INLINE;     M: B@R+   B@R+ INLINE;    M: OVER   over INLINE;
M: B!A   B!A INLINE;     M: B!R+   B!R+ INLINE;    M: DROP   drop INLINE;
M: @A+   @A+ INLINE;     M: INVERT   com INLINE;


\ Extra inline ops

M: NEGATE   T T 0 # RSB, INLINE;
M: -   R0 S POP,  T R0 T SUB, INLINE;
M: OR   R0 S POP,  T R0 T ORR, INLINE;
M: LSHIFT   R0 S POP,  T R0 T LSL MOV, INLINE;
M: RSHIFT   R0 S POP,  T R0 T LSR MOV, INLINE;


\ Utility words

M: EXECUTE   R0 T MOV,  T S POP,  PC R0 MOV, M;
M: SWAP   R0 S POP,  T S PUSH,  T R0 MOV,  ret M;
M: MIN   R0 S POP,  T R0 CMP,  T R0 GT MOV,  ret M;

M: EMIT   1 0 swi" OS_WriteC"  ret M;
M: SPACE   0 0 HEX 120 DECIMAL swi  ret M;
M: CR   0 0 swi" OS_NewLine" ret M;
M: BYE   0 0 swi" OS_NewLine"  0 lit  1 0 swi" OS_Exit" M;


\ Number I/O

CREATE BASE  10 ,
M: >DIGIT   ( n -- char )   -10 lit plus -if,  CHAR 0 10 + lit plus
   else, CHAR A lit plus  then,  ret M;
M: U.   col  A! -1 lit A@  $ BASE @A A!  begin,  A@ $ U/MOD  0until,  drop
   begin,  $ >DIGIT $ EMIT  -until,  drop  $ SPACE  scol M;
M: .   col  -if,  CHAR - lit $ EMIT  negate  then,  $ U.  scol M;

M: SIGN   ( start -- sign start' )   A!  B@A  CHAR - lit -or if,  drop
   1 lit A@  else, com A@ 1 lit plus  then,  ret M;
M: <DIGIT   ( char -- n )   CHAR A NEGATE lit plus -if,
   CHAR A CHAR 0 - lit plus  else, 10 lit plus  then,  ret M;
M: DIGIT?   ( n -- f )   col  $ BASE @A  1 lit minus  over minus  -or
   31 lit rshift  1 lit  -or  scol M;
M: NUMBER   ( start length -- n f )   col  over $ SIGN  push push plus
   A! pop A@  0 lit  dup begin, drop  r$ BASE @A mul  B@R+ r$ <DIGIT
   dup r$ DIGIT? if,  drop plus  over pop dup push  -or SWAP 0until,
   drop pop drop A!  drop A@ mul  -1 lit  scol  else,  drop drop drop drop
   pop drop  0 lit  scol  DROP M;


\ Terminal I/O

M: TYPE   col  push A! pop  begin,  B@A+ $ EMIT  -1 lit plus  0until, drop
   scol M;

CREATE TIB  SWAP DUP HERE >-< 8 -  OVER @ OR  SWAP !  0 ,
CREATE #TIB  0 ,
CREATE >IN  0 ,
M: REFILL   col  255 lit  32 lit  TIB-SIZE lit  $ TIB @A  dup $ >IN !A
   4 2 swi" OS_ReadLine"  drop  $ #TIB !A  scol M;

M: SENTINEL   ( char -- )   col  $ TIB @A  $ #TIB @A plus  A! B!A
   scol M;
M: SKIP   ( delim ptr -- ptr' )   A!  dup begin, drop  B@A+ over  -or until,
   drop drop  -1 lit A@ plus  ret M;
M: PARSE   ( delim ptr -- ptr' )   A!  dup begin, drop  B@A+ over -or 0until,
   drop drop  -1 lit A@ plus  ret M;
M: WORD   ( char -- start len )   col  dup com $ SENTINEL  dup  $ >IN @A
   $ SKIP  over $ SENTINEL  $ SWAP  over $ PARSE  dup $ >IN !A
   over minus  scol M;


\ Dictionary

M: >NAME   4 lit plus  ret M;
M: >CODE   12 lit plus  ret M;
CREATE THREADS  THREADS <ORG ,
M: THREAD   ( name -- 'thread )   col  dup 3 lit and 2mul 2mul
   A! 4 lit rshift  48 lit and  A@ or  $ THREADS @A plus  scol M;
M: NAME   ( start length -- name )   col  A! A@  3 lit $ MIN  A@ push push
   A! pop pop  A@ push  1 lit A!  dup begin, drop  B@R+  8 lit  A@
   mul lshift or  over  A@ dup  1 lit plus A!  -or 0until,  drop pop drop
   A! drop A@  scol M;
M: FIND   ( name -- xt | FALSE )  col  dup $ THREAD  begin, A! @A while,
   over over $ >NAME A! @A -or if,  drop -ROT repeat,  A! drop A@ scol  then,
   drop A! drop A@  $ >CODE  scol M;
M: '   col  BL lit $ WORD $ NAME $ FIND  scol M;


\ Compiler

CREATE DP  HERE 0 ,
CREATE STATE  1 ,   \ N.B.  0 = compiling, 1 = interpreting
M: IMMEDIATE?   ( xt -- f )   -1 lit plus A!  B@A  ret M;

M: SYNCHRONIZE   ( a-addr -- )   dup 1 lit 3 0 swi" XOS_SynchroniseCodeAreas"
   ret M;
M: CODE!   ( x -- )   ( address in A )   col  A@ push !A pop  $ SYNCHRONIZE
   scol M;
M: ,   col  $ DP @A  dup 4 lit plus !A  A! $ CODE!  scol M;
M: INLINE   ( xt -- )   HERE TO 'INLINE  col  dup -4 lit plus A!  B@A over
   plus  A! push A@  dup begin, drop  @R+ r$ ,  dup pop dup push -or 0until,
   drop drop pop drop  scol M;
M: [   col  1 lit $ STATE !A  scol M; IMMEDIATE
M: ]   col  0 lit $ STATE !A  scol M;

HEX
M: >BRANCH   ( from to -- offset )   A! negate A@ plus  2 lit rshift
   -2 lit plus  FFFFFF lit and  ret M;
M: BRANCH   col  push push  r$ DP @A pop  r$ >BRANCH  pop or  $ ,  scol M;
M: BRANCH,   col  EA000000 lit  $ BRANCH  scol M;
M: COMPILE,    HERE TO 'COMPILE,  col  EB000000 lit  $ BRANCH  scol M;
DECIMAL

M: (LITERAL)   pop  dup 4 lit plus push  HEX 03FFFFFF DECIMAL lit and A! @A
   ret M;
M: LITERAL   col  POSTPONE (LITERAL)  $ ,  scol M;

M: CODE   ( name )   col  BL lit $ WORD $ NAME  $ DP @A  over $ THREAD
   dup A!  @A $ ,  A! !A  $ ,  $ DP @A  0 lit $ ,  $ ]  scol M;
M: END-CODE   col  drop  $ [  scol M; IMMEDIATE
M: :   ( name )   col  $ CODE  POSTPONE col  scol M;
M: ;   col  POSTPONE scol  $ END-CODE  scol M; IMMEDIATE
M: CREATE   ( name )   col  $ : drop  POSTPONE (CREATE)  $ END-CODE  scol M;


\ Control structures

HEX
M: IF,   col  E35A0000 lit $ ,  0A000000 lit $ BRANCH  scol M;
M: -IF,   col  E31A0102 lit $ ,  0A000000 lit $ BRANCH  scol M;
DECIMAL

M: AHEAD   col  $ DP @A  dup 8 lit plus  $ BRANCH,  scol M; IMMEDIATE
M: IF   col  $ DP @A  4 lit plus  dup 8 lit plus  $ IF,  scol M; IMMEDIATE
M: -IF   col  $ DP @A  4 lit plus  dup 8 lit plus  $ -IF,  scol M; IMMEDIATE
M: THEN   col  dup A! @A  over $ DP @A $ >BRANCH  or  push A! pop  $ CODE!
   scol M; IMMEDIATE
M: ELSE   col  $ AHEAD  $ SWAP  $ THEN  scol M; IMMEDIATE

M: BEGIN   col  $ DP @A  scol M; IMMEDIATE
M: AGAIN   col  $ BRANCH,  scol M; IMMEDIATE
M: WHILE   col  $ IF $ SWAP  scol M; IMMEDIATE
M: -WHILE   col  $ -IF $ SWAP  scol M; IMMEDIATE
M: REPEAT   col  $ AGAIN $ THEN  scol M; IMMEDIATE
M: UNTIL   col  $ IF,  scol M; IMMEDIATE
M: -UNTIL   col  $ -IF,  scol M; IMMEDIATE
M: 0UNTIL   col  $ WHILE $ REPEAT  scol M; IMMEDIATE
M: +UNTIL   col  $ -WHILE $ REPEAT  scol M; IMMEDIATE


\ OS access

HEX
M: ONES   ( n -- 2^n-1 )   A! 1 lit A@  lshift  -1 lit plus  ret M;
M: SWI   col  E52CA004 lit $ ,  push push  if,  r$ ONES  E8BC0000 lit or
   r$ ,  else, drop  then,  pop pop  EF000000 lit or  $ ,  if,  $ ONES
   E92C0000 lit or  $ ,  then,  E49CA004 lit $ ,  scol M; IMMEDIATE
DECIMAL


\ Interpreter

M: COMPILE   ( xt -- )   col  dup -4 lit plus A!  B@A if,  drop  $ INLINE
   else, drop  $ COMPILE,  then,  scol M;
M: INTERPRET
   col
   begin,
      BL lit $ WORD while,
      over over
      $ NAME $ FIND if,
         A! drop drop A@
         dup $ IMMEDIATE?  $ STATE @A  or if,
            drop $ EXECUTE
         else,
            drop $ COMPILE
         then,
      else,
         drop  $ NUMBER if,
            drop  $ STATE @A if,
               drop
            else,
               drop  $ LITERAL
            then,
         else,
            CHAR ? lit $ EMIT  $ CR  IMAGE else
         then,
      then,
   repeat,
   drop drop  scol M;

M: QUIT
   IMAGE else   \ branch to start of image to reset stacks
   ROT then,    \ resolve initial branch
   1 lit  $ STATE !A
   begin,
      $ CR
      $ REFILL
      CHAR > lit $ EMIT  $ SPACE
      $ INTERPRET
   again, M;

HERE <ORG  SWAP !



\ Relocate and save image

HERE IMAGE - TO LENGTH

M0  IMAGE TO M0  RELOCATE  TO M0   \ relocate image
S" mf32"  IMAGE LENGTH 2OVER SAVE-FILE   \ write object module
HEX FF8 DECIMAL -ROT TYPE-FILE   \ set filetype to Absolute



\ Print out statistics

CR LENGTH CELL/ .  .( cells total )
CR CODE-CELLS ?  .( code cells )
CR MF-OPS ?  .( Machine Forth ops )
CR WORDS ?  .( definitions )



>SWITCH<   \ switch back to original definitions
KERNEL PREVIOUS DEFINITIONS   \ restore original order
